#!/usr/bin/env python
'''
Lucas-Kanade tracker
====================

Lucas-Kanade sparse optical flow demo. Uses goodFeaturesToTrack
for track initialization and back-tracking for match verification
between frames.

Usage
-----
lk_track.py [<video_source>]


Keys
----
ESC - exit
'''

# Python 2/3 compatibility
from __future__ import print_function

# 导入所需库
import numpy as np
import cv2 as cv

# 导入其他模块
import video
from common import anorm2, draw_str

# 设置Lucas-Kanade光流跟踪参数
lk_params = dict( winSize  = (15, 15),
                  maxLevel = 2,
                  criteria = (cv.TERM_CRITERIA_EPS | cv.TERM_CRITERIA_COUNT, 10, 0.03))

# 设置角点检测参数
feature_params = dict( maxCorners = 500,
                       qualityLevel = 0.3,
                       minDistance = 7,
                       blockSize = 7 )

class App:
    def __init__(self, video_src):
        # 初始���相关参数
        self.track_len = 10
        self.detect_interval = 5
        self.tracks = []
        # 创建视频捕获对象
        self.cam = video.create_capture(video_src)
        self.frame_idx = 0
        # 创建全屏窗口
        cv.namedWindow('lk_track',cv.WND_PROP_FULLSCREEN)
        cv.setWindowProperty('lk_track', cv.WND_PROP_FULLSCREEN, cv.WINDOW_FULLSCREEN)

    def run(self):
        while True:
            # 读取一帧图像
            _ret, frame = self.cam.read()
            # 转为灰度图像
            frame_gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
            vis = frame.copy()

            # 如果有跟踪点
            if len(self.tracks) > 0:
                # 获取前后帧图像
                img0, img1 = self.prev_gray, frame_gray
                # 获取跟踪点坐标
                p0 = np.float32([tr[-1] for tr in self.tracks]).reshape(-1, 1, 2)
                # 计算光流
                p1, _st, _err = cv.calcOpticalFlowPyrLK(img0, img1, p0, None, **lk_params)
                # 反向计算光流
                p0r, _st, _err = cv.calcOpticalFlowPyrLK(img1, img0, p1, None, **lk_params)
                # 计算反向光流误差
                d = abs(p0-p0r).reshape(-1, 2).max(-1)
                # 筛选出误差小于1的跟踪点
                good = d < 1
                new_tracks = []
                for tr, (x, y), good_flag in zip(self.tracks, p1.reshape(-1, 2), good):
                    # 如果误差过大，放弃该跟踪点
                    if not good_flag:
                        continue
                    # 添加新的跟踪点
                    tr.append((x, y))
                    # 如果跟踪点过多，删除旧的跟踪点
                    if len(tr) > self.track_len:
                        del tr[0]
                    # 添加新的跟踪路径
                    new_tracks.append(tr)
                    # 在图像上绘制跟踪点
                    cv.circle(vis, (int(x), int(y)), 2, (0, 255, 0), -1)
                # 更新跟踪路径
                self.tracks = new_tracks
                # 在图像上绘制跟踪路径
                cv.polylines(vis, [np.int32(tr) for tr in self.tracks], False, (0, 255, 0))
                # 在图像上显示跟踪点数目
                draw_str(vis, (20, 20), 'track count: %d' % len(self.tracks))

            # 在固定帧数间隔后检测新的特征点
            if self.frame_idx % self.detect_interval == 0:
                # 创建掩码
                mask = np.zeros_like(frame_gray)
                mask[:] = 255
                # 对已有跟踪点周围区域进行掩蔽
                for x, y in [np.int32(tr[-1]) for tr in self.tracks]:
                    cv.circle(mask, (x, y), 5, 0, -1)
                # 检测新的特征点
                p = cv.goodFeaturesToTrack(frame_gray, mask = mask, **feature_params)
                # 如果检测到特征点，添加到跟踪点中
                if p is not None:
                    for x, y in np.float32(p).reshape(-1, 2):
                        self.tracks.append([(x, y)])


            # 更新帧数和前一帧图像
            self.frame_idx += 1
            self.prev_gray = frame_gray
            # 显示图像
            cv.imshow('lk_track', vis)

            # 按下ESC键退出
            ch = cv.waitKey(1)
            if ch == 27:
                break

def main():
    import sys
    try:
        video_src = sys.argv[1]
    except:
        video_src = 0

    # 运行程序
    App(video_src).run()
    print('Done')


if __name__ == '__main__':
    print(__doc__)
    main()
    # 关闭所有窗口
    cv.destroyAllWindows()
